/*
 * Simulation of dining philosophers problem.
 *
 * Program requires the following command-line arguments:
 *   number of think/hungry/eat cycles per philosopher
 *   maximum time to think (milliseconds)
 *   maximum time to eat (milliseconds)
 *   optional seed for generating delays
 *
 * Code is designed to allow different implementations of parts involving
 * synchronization.
 */
#define _XOPEN_SOURCE 500 /* may be needed to get usleep() in unistd.h */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

void * philosopher_fcn(void *thread_arg);

/* global variables for runtime parameters */
int num_cycles;
int max_think_delay;
int max_eat_delay;

#include "getint.h"
#include "timer.h"
#include "test-dining-philosophers.h"
#include "dining-philosophers.h"

/* ---- main program ---- */

int main(int argc, char * argv[]) {

    char *usage_msg = "parameters:\n"
            "\tcycles\n"
            "\tthink_delay (msec)\n"
            "\teat_delay (msec)\n"
            "\tseed for random delays (optional)\n";

    if (argc < 4) {
        fprintf(stderr, usage_msg);
        exit(EXIT_FAILURE);
    }
    get_positive_int_or_exit(usage_msg, argv[1], &num_cycles);
    get_positive_int_or_exit(usage_msg, argv[2], &max_think_delay);
    get_positive_int_or_exit(usage_msg, argv[3], &max_eat_delay);
    if (argc > 4) {
        int seed;
        get_positive_int_or_exit(usage_msg, argv[4], &seed);
        srand(seed);
    }

    /* initialize timer */
    init_start_time();

    /* initialize for synchronization */
    init_synch();

    /* set up IDs for threads */
    int philosopherIDs[5];
    for (int i = 0; i < 5; ++i)
        philosopherIDs[i] = i;

    /* start threads */
    pthread_t philosophers[5];
    for (int i = 0; i < 5; ++i) 
        pthread_create(&philosophers[i], NULL, philosopher_fcn, 
                (void *) &philosopherIDs[i]);

    /* wait for all threads to complete */
    for (int i = 0; i < 5; ++i) 
        pthread_join(philosophers[i], NULL);

    /* clean up and exit */
    end_synch();
    return EXIT_SUCCESS;
}

/* ---- code to be executed by each philosopher ---- */

void * philosopher_fcn(void * thread_arg) {
    int myID = * (int *) thread_arg;

    for (int i = 0; i < num_cycles; ++i) {
        printf("at time %ld philosopher %d thinking\n", elapsed_time(), myID);
        int delay = (rand() % max_think_delay) + 1;
        usleep(delay * 1000);
        printf("at time %ld philosopher %d hungry\n", elapsed_time(), myID);
        simulate_eating_with_synch(myID);
    }

    pthread_exit((void* ) NULL);
}

/* ---- simulate eating ---- */

void simulate_eating(const int philosopherID) {
    int delay = (rand() % max_eat_delay) + 1;
    printf("at time %ld philosopher %d eating\n", elapsed_time(), philosopherID);
    usleep(delay * 1000);
}

/* ---- utility functions ---- */

int left(const int philosopherID) {
    return (philosopherID + 4) % 5;
}
int right(const int philosopherID) {
    return (philosopherID + 1) % 5;
}
void print_states(const char *msg, const pstate_t states[5]) {
    printf("%s ", msg);
    for (int i = 0; i < 5; ++i) {
        putchar(' ');
        switch(states[i]) {
            case thinking:
                putchar('t');
                break;
            case hungry:
                putchar('h');
                break;
            case eating:
                putchar('e');
                break;
        }
    }
    putchar('\n');
}

